from __future__ import annotations

from typing import TYPE_CHECKING

import click
from git import Repo

from semantic_release.cli.util import noop_report
from semantic_release.globals import logger
from semantic_release.hvcs.remote_hvcs_base import RemoteHvcsBase
from semantic_release.version.algorithm import tags_and_versions

if TYPE_CHECKING:  # pragma: no cover
    from semantic_release.cli.cli_context import CliContextObj


def publish_distributions(
    tag: str,
    hvcs_client: RemoteHvcsBase,
    dist_glob_patterns: tuple[str, ...],
    noop: bool = False,
) -> None:
    if noop:
        noop_report(
            str.join(
                " ",
                [
                    "would have uploaded files matching any of the globs",
                    str.join(", ", [repr(g) for g in dist_glob_patterns]),
                    "to a remote VCS release, if supported",
                ],
            )
        )
        return

    logger.info("Uploading distributions to release")
    for pattern in dist_glob_patterns:
        hvcs_client.upload_dists(tag=tag, dist_glob=pattern)  # type: ignore[attr-defined]


@click.command(
    short_help="Publish distributions to VCS Releases",
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
@click.option(
    "--tag",
    "tag",
    help="The tag associated with the release to publish to",
    default="latest",
)
@click.pass_obj
def publish(cli_ctx: CliContextObj, tag: str) -> None:
    """Build and publish a distribution to a VCS release."""
    ctx = click.get_current_context()
    runtime = cli_ctx.runtime_ctx
    hvcs_client = runtime.hvcs_client
    translator = runtime.version_translator
    dist_glob_patterns = runtime.dist_glob_patterns

    with Repo(str(runtime.repo_dir)) as git_repo:
        repo_tags = git_repo.tags

    if tag == "latest":
        try:
            tag = str(tags_and_versions(repo_tags, translator)[0][0])
        except IndexError:
            click.echo(
                str.join(
                    " ",
                    [
                        "No tags found with format",
                        repr(translator.tag_format),
                        "couldn't identify latest version",
                    ],
                ),
                err=True,
            )
            ctx.exit(1)

    if tag not in {tag.name for tag in repo_tags}:
        click.echo(f"Tag '{tag}' not found in local repository!", err=True)
        ctx.exit(1)

    if not isinstance(hvcs_client, RemoteHvcsBase):
        click.echo(
            "Remote does not support artifact upload. Exiting with no action taken...",
            err=True,
        )
        return

    publish_distributions(
        tag=tag,
        hvcs_client=hvcs_client,
        dist_glob_patterns=dist_glob_patterns,
        noop=runtime.global_cli_options.noop,
    )



from __future__ import annotations

from re import VERBOSE, compile as regexp, escape as regex_escape
from typing import TYPE_CHECKING

from semantic_release.const import SEMVER_REGEX
from semantic_release.globals import logger
from semantic_release.helpers import check_tag_format
from semantic_release.version.version import Version

if TYPE_CHECKING:
    from re import Pattern


class VersionTranslator:
    """
    Class to handle translation from Git tags into their corresponding Version
    instances.
    """

    _VERSION_REGEX = SEMVER_REGEX

    @classmethod
    def _invert_tag_format_to_re(cls, tag_format: str) -> Pattern[str]:
        r"""
        Unpick the "tag_format" format string and create a regex which can be used to
        convert a tag to a version string.

        The following relationship should always hold true:
        >>> version = "1.2.3-anything.1+at_all.1234"  # doesn't matter
        >>> tag_format = "v-anything_{version}_at-all"  # doesn't matter
        >>> inverted_format = VersionTranslator._invert_tag_format_to_re(tag_format)
        >>> tag = tag_format.format(version=version)
        >>> m = inverted_format.match(tag)
        >>> assert m is not None
        >>> assert m.expand(r"\g<version>") == version
        """
        pat = regexp(
            regex_escape(tag_format).replace(
                regex_escape(r"{version}"), r"(?P<version>.+)"
            ),
            flags=VERBOSE,
        )
        logger.debug("inverted tag_format %r to %r", tag_format, pat.pattern)
        return pat

    def __init__(
        self,
        tag_format: str = "v{version}",
        prerelease_token: str = "rc",  # noqa: S107
        add_partial_tags: bool = False,
    ) -> None:
        check_tag_format(tag_format)
        self.tag_format = tag_format
        self.prerelease_token = prerelease_token
        self.add_partial_tags = add_partial_tags
        self.from_tag_re = self._invert_tag_format_to_re(self.tag_format)
        self.partial_tag_re = regexp(
            regex_escape(tag_format).replace(
                regex_escape(r"{version}"), r"[0-9]+(\.(0|[1-9][0-9]*))?$"
            ),
            flags=VERBOSE,
        )

    def from_string(self, version_str: str) -> Version:
        """
        Return a Version instance from a string. Delegates directly to Version.parse,
        using the translator's own stored values for tag_format and prerelease
        """
        return Version.parse(
            version_str,
            tag_format=self.tag_format,
            prerelease_token=self.prerelease_token,
        )

    def from_tag(self, tag: str) -> Version | None:
        """
        Return a Version instance from a Git tag, if tag_format matches the format
        which would have generated the tag from a version. Otherwise return None.
        For example, a tag of 'v1.2.3' should be matched if `tag_format = 'v{version}`,
        but not if `tag_format = staging--v{version}`.
        """
        tag_match = self.from_tag_re.match(tag)
        if not tag_match:
            return None
        if self.add_partial_tags:
            partial_tag_match = self.partial_tag_re.match(tag)
            if partial_tag_match:
                return None
        raw_version_str = tag_match.group("version")
        return self.from_string(raw_version_str)

    def str_to_tag(self, version_str: str) -> str:
        """Formats a version string into a tag name"""
        return self.tag_format.format(version=version_str)

    def __repr__(self) -> str:
        return (
            f"{type(self).__qualname__}(tag_format={self.tag_format}, "
            f"prerelease_token={self.prerelease_token})"
        )



from __future__ import annotations

import json
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, cast

import pytest
import tomlkit
import yaml
from dotty_dict import Dotty
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.version.declarations.enum import VersionStampType

from tests.const import EXAMPLE_PROJECT_NAME, MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.repos.trunk_based_dev.repo_w_no_tags import (
    repo_w_no_tags_conventional_commits,
)
from tests.fixtures.repos.trunk_based_dev.repo_w_prereleases import (
    repo_w_trunk_only_n_prereleases_conventional_commits,
)
from tests.util import (
    assert_successful_exit_code,
    dynamic_python_import,
)

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import ExProjectDir, UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import BuiltRepoResult


VERSION_STAMP_CMD = [
    MAIN_PROG_NAME,
    VERSION_SUBCMD,
    "--no-commit",
    "--no-tag",
    "--skip-build",
    "--no-changelog",
]
"""Using the version command, prevent any action besides stamping the version"""


@pytest.mark.parametrize(
    "repo_result, expected_new_version",
    [
        (
            lazy_fixture(repo_w_trunk_only_n_prereleases_conventional_commits.__name__),
            "0.3.0",
        )
    ],
)
def test_version_only_stamp_version(
    repo_result: BuiltRepoResult,
    expected_new_version: str,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: MagicMock,
    example_pyproject_toml: Path,
    example_project_dir: ExProjectDir,
    pyproject_toml_file: Path,
) -> None:
    repo = repo_result["repo"]
    version_file = example_project_dir.joinpath(
        "src", EXAMPLE_PROJECT_NAME, "_version.py"
    )
    expected_changed_files = sorted(
        [
            str(pyproject_toml_file),
            str(version_file.relative_to(example_project_dir)),
        ]
    )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}
    version_py_before = dynamic_python_import(
        version_file, f"{EXAMPLE_PROJECT_NAME}._version"
    ).__version__

    pyproject_toml_before = tomlkit.loads(
        example_pyproject_toml.read_text(encoding="utf-8")
    )

    # Modify the pyproject.toml to remove the version so we can compare it later
    pyproject_toml_before.get("tool", {}).get("poetry", {}).pop("version")

    # Act (stamp the version but also create the changelog)
    cli_cmd = [*VERSION_STAMP_CMD, "--minor"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = set.difference(tags_after, tags_before)
    actual_staged_files = [
        # Make sure filepath uses os specific path separators
        str(Path(file))
        # Changed files should always be staged
        for file in cast("str", repo.git.diff(staged=True, name_only=True)).splitlines()
    ]
    pyproject_toml_after = tomlkit.loads(
        example_pyproject_toml.read_text(encoding="utf-8")
    )
    pyproj_version_after = (
        pyproject_toml_after.get("tool", {}).get("poetry", {}).pop("version")
    )

    # Load python module for reading the version (ensures the file is valid)
    version_py_after = dynamic_python_import(
        version_file, f"{EXAMPLE_PROJECT_NAME}._version"
    ).__version__

    # Evaluate (no release actions should be taken but version should be stamped from forced minor bump)
    assert_successful_exit_code(result, cli_cmd)

    assert head_sha_before == head_after.hexsha  # No commit should be made
    assert not tags_set_difference  # No tag should be created

    # no push as it should be turned off automatically
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0  # no vcs release creation occurred

    # Files that should receive version change
    assert expected_changed_files == actual_staged_files

    # Compare pyproject.toml
    assert pyproject_toml_before == pyproject_toml_after
    assert expected_new_version == pyproj_version_after

    # Compare _version.py
    assert expected_new_version == version_py_after
    assert version_py_before != version_py_after


# ============================================================================== #
#                     VERSION STAMP DIFFERENT CONTENT TYPES                      #
# ============================================================================== #


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_stamp_version_variables_python(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    example_project_dir: ExProjectDir,
) -> None:
    new_version = "1.0.0"
    target_file = example_project_dir.joinpath(
        "src", EXAMPLE_PROJECT_NAME, "_version.py"
    )

    # Set configuration to modify the python file
    update_pyproject_toml(
        "tool.semantic_release.version_variables",
        [f"{target_file.relative_to(example_project_dir)}:__version__"],
    )

    # Act
    cli_cmd = VERSION_STAMP_CMD
    result = run_cli(cli_cmd[1:])

    # Check the result
    assert_successful_exit_code(result, cli_cmd)

    # Load python module for reading the version (ensures the file is valid)
    version_py_after = dynamic_python_import(
        target_file, f"{EXAMPLE_PROJECT_NAME}._version"
    ).__version__

    # Check the version was updated
    assert new_version == version_py_after


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_stamp_version_toml(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    default_tag_format_str: str,
) -> None:
    orig_version = "0.0.0"
    new_version = "1.0.0"
    orig_release = default_tag_format_str.format(version=orig_version)
    new_release = default_tag_format_str.format(version=new_version)
    target_file = Path("example.toml")
    orig_toml = dedent(
        f"""\
        [package]
        name = "example"
        version = "{orig_version}"
        release = "{orig_release}"
        date-released = "1970-01-01"
        """
    )

    orig_toml_obj = Dotty(tomlkit.parse(orig_toml))

    # Write initial text in file
    target_file.write_text(orig_toml)

    # Set configuration to modify the yaml file
    update_pyproject_toml(
        "tool.semantic_release.version_toml",
        [
            f"{target_file}:package.version:{VersionStampType.NUMBER_FORMAT.value}",
            f"{target_file}:package.release:{VersionStampType.TAG_FORMAT.value}",
        ],
    )

    # Act
    cli_cmd = VERSION_STAMP_CMD
    result = run_cli(cli_cmd[1:])

    # Check the result
    assert_successful_exit_code(result, cli_cmd)

    # Read content
    resulting_toml_obj = Dotty(tomlkit.parse(target_file.read_text()))

    # Check the version was updated
    assert new_version == resulting_toml_obj["package.version"]
    assert new_release == resulting_toml_obj["package.release"]

    # Check the rest of the content is the same (by resetting the version & comparing)
    resulting_toml_obj["package.version"] = orig_version
    resulting_toml_obj["package.release"] = orig_release

    assert orig_toml_obj == resulting_toml_obj


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_stamp_version_variables_yaml(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
) -> None:
    orig_version = "0.0.0"
    new_version = "1.0.0"
    target_file = Path("example.yml")
    orig_yaml = dedent(
        f"""\
        ---
        package: example
        version: {orig_version}
        date-released: 1970-01-01
        """
    )
    # Write initial text in file
    target_file.write_text(orig_yaml)

    # Set configuration to modify the yaml file
    update_pyproject_toml(
        "tool.semantic_release.version_variables", [f"{target_file}:version"]
    )

    # Act
    cli_cmd = VERSION_STAMP_CMD
    result = run_cli(cli_cmd[1:])

    # Check the result
    assert_successful_exit_code(result, cli_cmd)

    # Read content
    resulting_yaml_obj = yaml.safe_load(target_file.read_text())

    # Check the version was updated
    assert new_version == resulting_yaml_obj["version"]

    # Check the rest of the content is the same (by resetting the version & comparing)
    resulting_yaml_obj["version"] = orig_version

    assert yaml.safe_load(orig_yaml) == resulting_yaml_obj


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_stamp_version_variables_yaml_cff(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
) -> None:
    """
    Given a yaml file with a top level version directive,
    When the version command is run,
    Then the version is updated in the file and the rest of the content is unchanged & parsable

    Based on https://github.com/python-semantic-release/python-semantic-release/issues/962
    """
    orig_version = "0.0.0"
    new_version = "1.0.0"
    target_file = Path("CITATION.cff")
    orig_yaml = dedent(
        f"""\
        ---
        cff-version: 1.2.0
        message: "If you use this software, please cite it as below."
        authors:
            - family-names: Doe
              given-names: Jon
              orcid: https://orcid.org/1234-6666-2222-5555
        title: "My Research Software"
        version: {orig_version}
        date-released: 1970-01-01
        """
    )
    # Write initial text in file
    target_file.write_text(orig_yaml)

    # Set configuration to modify the yaml file
    update_pyproject_toml(
        "tool.semantic_release.version_variables", [f"{target_file}:version"]
    )

    # Act
    cli_cmd = VERSION_STAMP_CMD
    result = run_cli(cli_cmd[1:])

    # Check the result
    assert_successful_exit_code(result, cli_cmd)

    # Read content
    resulting_yaml_obj = yaml.safe_load(target_file.read_text())

    # Check the version was updated
    assert new_version == resulting_yaml_obj["version"]

    # Check the rest of the content is the same (by resetting the version & comparing)
    resulting_yaml_obj["version"] = orig_version

    assert yaml.safe_load(orig_yaml) == resulting_yaml_obj


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_stamp_version_variables_json(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
) -> None:
    orig_version = "0.0.0"
    new_version = "1.0.0"
    target_file = Path("plugins.json")
    orig_json = {
        "id": "test-plugin",
        "version": orig_version,
        "meta": {
            "description": "Test plugin",
        },
    }
    # Write initial text in file
    target_file.write_text(json.dumps(orig_json, indent=4))

    # Set configuration to modify the json file
    update_pyproject_toml(
        "tool.semantic_release.version_variables", [f"{target_file}:version"]
    )

    # Act
    cli_cmd = VERSION_STAMP_CMD
    result = run_cli(cli_cmd[1:])

    # Check the result
    assert_successful_exit_code(result, cli_cmd)

    # Read content
    resulting_json_obj = json.loads(target_file.read_text())

    # Check the version was updated
    assert new_version == resulting_json_obj["version"]

    # Check the rest of the content is the same (by resetting the version & comparing)
    resulting_json_obj["version"] = orig_version

    assert orig_json == resulting_json_obj


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_stamp_version_variables_yaml_github_actions(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    default_tag_format_str: str,
) -> None:
    """
    Given a yaml file with github actions 'uses:' directives which use @vX.Y.Z version declarations,
    When a version is stamped and configured to stamp the version using the tag format,
    Then the file is updated with the new version in the tag format

    Based on https://github.com/python-semantic-release/python-semantic-release/issues/1156
    """
    orig_version = "0.0.0"
    new_version = "1.0.0"
    target_file = Path("combined.yml")
    action1_yaml_filepath = "my-org/my-actions/.github/workflows/action1.yml"
    action2_yaml_filepath = "my-org/my-actions/.github/workflows/action2.yml"
    orig_yaml = dedent(
        f"""\
        ---
        on:
          workflow_call:

        jobs:
          action1:
            uses: {action1_yaml_filepath}@{default_tag_format_str.format(version=orig_version)}
          action2:
            uses: {action2_yaml_filepath}@{default_tag_format_str.format(version=orig_version)}
        """
    )
    expected_action1_value = (
        f"{action1_yaml_filepath}@{default_tag_format_str.format(version=new_version)}"
    )
    expected_action2_value = (
        f"{action2_yaml_filepath}@{default_tag_format_str.format(version=new_version)}"
    )

    # Setup: Write initial text in file
    target_file.write_text(orig_yaml)

    # Setup: Set configuration to modify the yaml file
    update_pyproject_toml(
        "tool.semantic_release.version_variables",
        [
            f"{target_file}:{action1_yaml_filepath}:{VersionStampType.TAG_FORMAT.value}",
            f"{target_file}:{action2_yaml_filepath}:{VersionStampType.TAG_FORMAT.value}",
        ],
    )

    # Act
    cli_cmd = VERSION_STAMP_CMD
    result = run_cli(cli_cmd[1:])

    # Check the result
    assert_successful_exit_code(result, cli_cmd)

    # Read content
    resulting_yaml_obj = yaml.safe_load(target_file.read_text())

    # Check the version was updated
    assert expected_action1_value == resulting_yaml_obj["jobs"]["action1"]["uses"]
    assert expected_action2_value == resulting_yaml_obj["jobs"]["action2"]["uses"]

    # Check the rest of the content is the same (by setting the version & comparing)
    original_yaml_obj = yaml.safe_load(orig_yaml)
    original_yaml_obj["jobs"]["action1"]["uses"] = expected_action1_value
    original_yaml_obj["jobs"]["action2"]["uses"] = expected_action2_value

    assert original_yaml_obj == resulting_yaml_obj


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_stamp_version_variables_yaml_kustomization_container_spec(
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    default_tag_format_str: str,
) -> None:
    """
    Given a yaml file with directives that expect a vX.Y.Z version tag declarations,
    When a version is stamped and configured to stamp the version using the tag format,
    Then the file is updated with the new version in the tag format

    Based on https://github.com/python-semantic-release/python-semantic-release/issues/846
    """
    orig_version = "0.0.0"
    new_version = "1.0.0"
    target_file = Path("kustomization.yaml")
    orig_yaml = dedent(
        f"""\
        images:
          - name: repo/image
            newTag: {default_tag_format_str.format(version=orig_version)}
        """
    )
    expected_new_tag_value = default_tag_format_str.format(version=new_version)

    # Setup: Write initial text in file
    target_file.write_text(orig_yaml)

    # Setup: Set configuration to modify the yaml file
    update_pyproject_toml(
        "tool.semantic_release.version_variables",
        [
            f"{target_file}:newTag:{VersionStampType.TAG_FORMAT.value}",
        ],
    )

    # Act
    cli_cmd = VERSION_STAMP_CMD
    result = run_cli(cli_cmd[1:])

    # Check the result
    assert_successful_exit_code(result, cli_cmd)

    # Read content
    resulting_yaml_obj = yaml.safe_load(target_file.read_text())

    # Check the version was updated
    assert expected_new_tag_value == resulting_yaml_obj["images"][0]["newTag"]

    # Check the rest of the content is the same (by resetting the version & comparing)
    original_yaml_obj = yaml.safe_load(orig_yaml)
    resulting_yaml_obj["images"][0]["newTag"] = original_yaml_obj["images"][0]["newTag"]

    assert original_yaml_obj == resulting_yaml_obj



from __future__ import annotations

from datetime import timedelta
from itertools import count
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest
import tomlkit

# Limitation in pytest-lazy-fixture - see https://stackoverflow.com/a/69884019
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.commit_parser.conventional import ConventionalCommitParser
from semantic_release.commit_parser.emoji import EmojiCommitParser
from semantic_release.commit_parser.scipy import ScipyCommitParser

from tests.const import EXAMPLE_PROJECT_NAME, MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures import (
    conventional_chore_commits,
    conventional_major_commits,
    conventional_minor_commits,
    conventional_patch_commits,
    emoji_chore_commits,
    emoji_major_commits,
    emoji_minor_commits,
    emoji_patch_commits,
    repo_w_git_flow_w_alpha_prereleases_n_conventional_commits,
    repo_w_git_flow_w_alpha_prereleases_n_emoji_commits,
    repo_w_git_flow_w_alpha_prereleases_n_scipy_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits,
    repo_w_github_flow_w_feature_release_channel_conventional_commits,
    repo_w_initial_commit,
    repo_w_no_tags_conventional_commits,
    repo_w_no_tags_conventional_commits_w_zero_version,
    repo_w_no_tags_emoji_commits,
    repo_w_no_tags_scipy_commits,
    repo_w_trunk_only_conventional_commits,
    repo_w_trunk_only_emoji_commits,
    repo_w_trunk_only_n_prereleases_conventional_commits,
    repo_w_trunk_only_n_prereleases_emoji_commits,
    repo_w_trunk_only_n_prereleases_scipy_commits,
    repo_w_trunk_only_scipy_commits,
    scipy_chore_commits,
    scipy_major_commits,
    scipy_minor_commits,
    scipy_patch_commits,
)
from tests.util import (
    add_text_to_file,
    assert_successful_exit_code,
    dynamic_python_import,
    xdist_sort_hack,
)

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from requests_mock import Mocker

    from tests.conftest import GetStableDateNowFn, RunCliFn
    from tests.fixtures.example_project import (
        ExProjectDir,
        GetExpectedVersionPyFileContentFn,
        UpdatePyprojectTomlFn,
    )
    from tests.fixtures.git_repo import BuiltRepoResult


@pytest.mark.parametrize(
    "repo_result, cli_args, next_release_version",
    [
        *(
            (
                lazy_fixture(
                    repo_w_no_tags_conventional_commits_w_zero_version.__name__
                ),
                cli_args,
                next_release_version,
            )
            for cli_args, next_release_version in (
                # New build-metadata forces a new release
                (["--build-metadata", "build.12345"], "0.1.0+build.12345"),
                # Forced version bump
                (["--prerelease"], "0.0.0-rc.1"),
                (["--patch"], "0.0.1"),
                (["--minor"], "0.1.0"),
                (["--major"], "1.0.0"),
                # Forced version bump with --build-metadata
                (["--patch", "--build-metadata", "build.12345"], "0.0.1+build.12345"),
                # Forced version bump with --as-prerelease
                (["--prerelease", "--as-prerelease"], "0.0.0-rc.1"),
                (["--patch", "--as-prerelease"], "0.0.1-rc.1"),
                (["--minor", "--as-prerelease"], "0.1.0-rc.1"),
                (["--major", "--as-prerelease"], "1.0.0-rc.1"),
                # Forced version bump with --as-prerelease and modified --prerelease-token
                (
                    ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                    "0.0.1-beta.1",
                ),
                # Forced version bump with --as-prerelease and modified --prerelease-token
                # and --build-metadata
                (
                    [
                        "--patch",
                        "--as-prerelease",
                        "--prerelease-token",
                        "beta",
                        "--build-metadata",
                        "build.12345",
                    ],
                    "0.0.1-beta.1+build.12345",
                ),
            )
        ),
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                cli_args,
                expected_stdout,
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture_name, values in {
                repo_w_trunk_only_conventional_commits.__name__: [
                    # New build-metadata forces a new release
                    (["--build-metadata", "build.12345"], "0.1.1+build.12345"),
                    # Forced version bump
                    (["--prerelease"], "0.1.1-rc.1"),
                    (["--patch"], "0.1.2"),
                    (["--minor"], "0.2.0"),
                    (["--major"], "1.0.0"),
                    # Forced version bump with --build-metadata
                    (
                        ["--patch", "--build-metadata", "build.12345"],
                        "0.1.2+build.12345",
                    ),
                    # Forced version bump with --as-prerelease
                    (["--prerelease", "--as-prerelease"], "0.1.1-rc.1"),
                    (["--patch", "--as-prerelease"], "0.1.2-rc.1"),
                    (["--minor", "--as-prerelease"], "0.2.0-rc.1"),
                    (["--major", "--as-prerelease"], "1.0.0-rc.1"),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    (
                        ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                        "0.1.2-beta.1",
                    ),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    # and --build-metadata
                    (
                        [
                            "--patch",
                            "--as-prerelease",
                            "--prerelease-token",
                            "beta",
                            "--build-metadata",
                            "build.12345",
                        ],
                        "0.1.2-beta.1+build.12345",
                    ),
                ],
                repo_w_trunk_only_n_prereleases_conventional_commits.__name__: [
                    # New build-metadata forces a new release
                    (["--build-metadata", "build.12345"], "0.2.0+build.12345"),
                    # Forced version bump
                    # NOTE: There is already a 0.2.0-rc.1
                    (["--prerelease"], "0.2.0-rc.2"),
                    (["--patch"], "0.2.1"),
                    (["--minor"], "0.3.0"),
                    (["--major"], "1.0.0"),
                    # Forced version bump with --build-metadata
                    (
                        ["--patch", "--build-metadata", "build.12345"],
                        "0.2.1+build.12345",
                    ),
                    # Forced version bump with --as-prerelease
                    (["--prerelease", "--as-prerelease"], "0.2.0-rc.2"),
                    (["--patch", "--as-prerelease"], "0.2.1-rc.1"),
                    (["--minor", "--as-prerelease"], "0.3.0-rc.1"),
                    (["--major", "--as-prerelease"], "1.0.0-rc.1"),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    (
                        ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                        "0.2.1-beta.1",
                    ),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    # and --build-metadata
                    (
                        [
                            "--patch",
                            "--as-prerelease",
                            "--prerelease-token",
                            "beta",
                            "--build-metadata",
                            "build.12345",
                        ],
                        "0.2.1-beta.1+build.12345",
                    ),
                ],
                repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__: [
                    # New build-metadata forces a new release
                    (["--build-metadata", "build.12345"], "1.1.0+build.12345"),
                    # Forced version bump
                    (["--prerelease"], "1.1.0-rc.1"),
                    (["--patch"], "1.1.1"),
                    (["--minor"], "1.2.0"),
                    (["--major"], "2.0.0"),
                    # Forced version bump with --build-metadata
                    (
                        ["--patch", "--build-metadata", "build.12345"],
                        "1.1.1+build.12345",
                    ),
                    # Forced version bump with --as-prerelease
                    (["--prerelease", "--as-prerelease"], "1.1.0-rc.1"),
                    (["--patch", "--as-prerelease"], "1.1.1-rc.1"),
                    (["--minor", "--as-prerelease"], "1.2.0-rc.1"),
                    (["--major", "--as-prerelease"], "2.0.0-rc.1"),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    (
                        ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                        "1.1.1-beta.1",
                    ),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    # and --build-metadata
                    (
                        [
                            "--patch",
                            "--as-prerelease",
                            "--prerelease-token",
                            "beta",
                            "--build-metadata",
                            "build.12345",
                        ],
                        "1.1.1-beta.1+build.12345",
                    ),
                ],
                repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__: [
                    # New build-metadata forces a new release
                    (["--build-metadata", "build.12345"], "1.2.0-alpha.2+build.12345"),
                    # Forced version bump
                    (["--prerelease"], "1.2.0-alpha.3"),
                    (["--patch"], "1.2.1"),
                    (["--minor"], "1.3.0"),
                    (["--major"], "2.0.0"),
                    # Forced version bump with --build-metadata
                    (
                        ["--patch", "--build-metadata", "build.12345"],
                        "1.2.1+build.12345",
                    ),
                    # Forced version bump with --as-prerelease
                    (["--prerelease", "--as-prerelease"], "1.2.0-alpha.3"),
                    (["--patch", "--as-prerelease"], "1.2.1-alpha.1"),
                    (["--minor", "--as-prerelease"], "1.3.0-alpha.1"),
                    (["--major", "--as-prerelease"], "2.0.0-alpha.1"),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    (
                        ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                        "1.2.1-beta.1",
                    ),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    # and --build-metadata
                    (
                        [
                            "--patch",
                            "--as-prerelease",
                            "--prerelease-token",
                            "beta",
                            "--build-metadata",
                            "build.12345",
                        ],
                        "1.2.1-beta.1+build.12345",
                    ),
                ],
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__: [
                    # New build-metadata forces a new release
                    (["--build-metadata", "build.12345"], "1.1.0+build.12345"),
                    # Forced version bump
                    (["--prerelease"], "1.1.0-rc.3"),
                    (["--patch"], "1.1.1"),
                    (["--minor"], "1.2.0"),
                    (["--major"], "2.0.0"),
                    # Forced version bump with --build-metadata
                    (
                        ["--patch", "--build-metadata", "build.12345"],
                        "1.1.1+build.12345",
                    ),
                    # Forced version bump with --as-prerelease
                    (["--prerelease", "--as-prerelease"], "1.1.0-rc.3"),
                    (["--patch", "--as-prerelease"], "1.1.1-rc.1"),
                    (["--minor", "--as-prerelease"], "1.2.0-rc.1"),
                    (["--major", "--as-prerelease"], "2.0.0-rc.1"),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    (
                        ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                        "1.1.1-beta.1",
                    ),
                    # Forced version bump with --as-prerelease and modified --prerelease-token
                    # and --build-metadata
                    (
                        [
                            "--patch",
                            "--as-prerelease",
                            "--prerelease-token",
                            "beta",
                            "--build-metadata",
                            "build.12345",
                        ],
                        "1.1.1-beta.1+build.12345",
                    ),
                ],
            }.items()
            for (cli_args, expected_stdout) in values
        ],
    ],
)
def test_version_force_level(
    repo_result: BuiltRepoResult,
    cli_args: list[str],
    next_release_version: str,
    example_project_dir: ExProjectDir,
    example_pyproject_toml: Path,
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    pyproject_toml_file: Path,
    changelog_md_file: Path,
    get_expected_version_py_file_content: GetExpectedVersionPyFileContentFn,
):
    # Force clean directory state before test (needed for the repo_w_no_tags)
    repo = repo_result["repo"]
    repo.git.reset("HEAD", hard=True)

    version_file = example_project_dir.joinpath(
        "src", EXAMPLE_PROJECT_NAME, "_version.py"
    )

    expected_changed_files = sorted(
        [
            str(changelog_md_file),
            str(pyproject_toml_file),
            str(version_file.relative_to(example_project_dir)),
        ]
    )

    expected_version_py_content = get_expected_version_py_file_content(
        next_release_version
    )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    pyproject_toml_before = tomlkit.loads(
        example_pyproject_toml.read_text(encoding="utf-8")
    )

    # Modify the pyproject.toml to remove the version so we can compare it later
    pyproject_toml_before.get("tool", {}).get("poetry").pop("version")  # type: ignore[attr-defined]

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *cli_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))
    differing_files = [
        # Make sure filepath uses os specific path separators
        str(Path(file))
        for file in str(repo.git.diff("HEAD", "HEAD~1", name_only=True)).splitlines()
    ]
    pyproject_toml_after = tomlkit.loads(
        example_pyproject_toml.read_text(encoding="utf-8")
    )
    pyproj_version_after = (
        pyproject_toml_after.get("tool", {}).get("poetry", {}).pop("version")
    )

    # Load python module for reading the version (ensures the file is valid)
    actual_version_py_content = version_file.read_text()

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred

    # Changelog already reflects changes this should introduce
    assert expected_changed_files == differing_files

    # Compare pyproject.toml
    assert pyproject_toml_before == pyproject_toml_after
    assert next_release_version == pyproj_version_after

    # Compare _version.py
    assert expected_version_py_content == actual_version_py_content

    # Verify content is parsable & importable
    dynamic_version = dynamic_python_import(
        version_file, f"{EXAMPLE_PROJECT_NAME}._version"
    ).__version__

    assert next_release_version == dynamic_version


# NOTE: There is a bit of a corner-case where if we are not doing a
# prerelease, we will get a full version based on already-released commits.
# So for example, commits that wouldn't trigger a release on a prerelease branch
# won't trigger a release if prerelease=true; however, when commits included in a
# prerelease branch are merged to a release branch, prerelease=False - so a feat commit
# which previously triggered a prerelease on a branch will subsequently trigger a full
# release when merged to a full release branch where prerelease=False.
# For this reason a couple of these test cases predict a new version even when the
# commits being added here don't induce a version bump.
@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            (
                # Default case should be a minor bump since last full release was 1.1.1
                # last tag is a prerelease 1.2.0-rc.2
                lazy_fixture(
                    repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__
                ),
                lazy_fixture(conventional_minor_commits.__name__),
                False,
                "alpha",
                "1.2.0",
                "main",
            ),
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    [] if commit_messages is None else lazy_fixture(commit_messages),
                    prerelease,
                    prerelease_token,
                    expected_new_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_git_flow is currently 1.2.0-alpha.2
                    # The last full release version was 1.1.1, so it's had a minor
                    # prerelease
                    (
                        repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                        "alpha",
                    ): [
                        (conventional_patch_commits.__name__, False, "1.1.2", None),
                        (
                            conventional_patch_commits.__name__,
                            True,
                            "1.1.2-alpha.1",
                            None,
                        ),
                        (
                            conventional_minor_commits.__name__,
                            True,
                            "1.2.0-alpha.3",
                            "feat/feature-4",  # branch
                        ),
                        (conventional_major_commits.__name__, False, "2.0.0", None),
                        (
                            conventional_major_commits.__name__,
                            True,
                            "2.0.0-alpha.1",
                            None,
                        ),
                    ],
                    # Latest version for repo_with_git_flow_and_release_channels is
                    # currently 1.1.0
                    (
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
                        "alpha",
                    ): [
                        (conventional_patch_commits.__name__, False, "1.1.1", None),
                        (
                            conventional_patch_commits.__name__,
                            True,
                            "1.1.1-alpha.1",
                            None,
                        ),
                        (conventional_minor_commits.__name__, False, "1.2.0", None),
                        (
                            conventional_minor_commits.__name__,
                            True,
                            "1.2.0-alpha.1",
                            None,
                        ),
                        (conventional_major_commits.__name__, False, "2.0.0", None),
                        (
                            conventional_major_commits.__name__,
                            True,
                            "2.0.0-alpha.1",
                            None,
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    expected_new_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ]
    ),
)
# TODO: add a github flow test case
def test_version_next_greater_than_version_one_conventional(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    run_cli: RunCliFn,
    file_in_repo: str,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    [] if commit_messages is None else lazy_fixture(commit_messages),
                    prerelease,
                    prerelease_token,
                    expected_new_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_git_flow is currently 1.2.0-alpha.2
                    # The last full release version was 1.1.1, so it's had a minor
                    # prerelease
                    (
                        repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                        "alpha",
                    ): [
                        *(
                            (commits, True, "1.2.0-alpha.2", "feat/feature-4")
                            for commits in (
                                None,
                                conventional_chore_commits.__name__,
                            )
                        ),
                        *(
                            (commits, False, "1.1.1", None)
                            for commits in (
                                None,
                                conventional_chore_commits.__name__,
                            )
                        ),
                    ],
                    # Latest version for repo_with_git_flow_and_release_channels is
                    # currently 1.1.0
                    (
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
                        "alpha",
                    ): [
                        *(
                            (commits, prerelease, "1.1.0", None)
                            for prerelease in (True, False)
                            for commits in (
                                None,
                                conventional_chore_commits.__name__,
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    expected_new_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ]
    ),
)
def test_version_next_greater_than_version_one_no_bump_conventional(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    run_cli: RunCliFn,
    file_in_repo: str,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (no release actions should have occurred when no bump)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # No commit has been made
    assert head_sha_before == head_after.hexsha
    assert len(tags_set_difference) == 0  # No tag created
    assert mocked_git_fetch.call_count == 0  # no git fetch called
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            pytest.param(
                lazy_fixture(repo_fixture_name),
                [] if commit_messages is None else lazy_fixture(commit_messages),
                prerelease,
                prerelease_token,
                expected_new_version,
                "main" if branch_name is None else branch_name,
                marks=pytest.mark.comprehensive,
            )
            for (repo_fixture_name, prerelease_token), values in {
                # Latest version for repo_with_git_flow is currently 1.2.0-alpha.2
                # The last full release version was 1.1.1, so it's had a minor
                # prerelease
                (
                    repo_w_git_flow_w_alpha_prereleases_n_emoji_commits.__name__,
                    "alpha",
                ): [
                    (emoji_patch_commits.__name__, False, "1.1.2", None),
                    (
                        emoji_patch_commits.__name__,
                        True,
                        "1.1.2-alpha.1",
                        None,
                    ),
                    (
                        emoji_minor_commits.__name__,
                        False,
                        "1.2.0",
                        None,
                    ),
                    (
                        emoji_minor_commits.__name__,
                        True,
                        "1.2.0-alpha.3",
                        "feat/feature-4",  # branch
                    ),
                    (emoji_major_commits.__name__, False, "2.0.0", None),
                    (
                        emoji_major_commits.__name__,
                        True,
                        "2.0.0-alpha.1",
                        None,
                    ),
                ],
                # Latest version for repo_with_git_flow_and_release_channels is
                # currently 1.1.0
                (
                    repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits.__name__,
                    "alpha",
                ): [
                    (emoji_patch_commits.__name__, False, "1.1.1", None),
                    (
                        emoji_patch_commits.__name__,
                        True,
                        "1.1.1-alpha.1",
                        None,
                    ),
                    (emoji_minor_commits.__name__, False, "1.2.0", None),
                    (
                        emoji_minor_commits.__name__,
                        True,
                        "1.2.0-alpha.1",
                        None,
                    ),
                    (emoji_major_commits.__name__, False, "2.0.0", None),
                    (
                        emoji_major_commits.__name__,
                        True,
                        "2.0.0-alpha.1",
                        None,
                    ),
                ],
            }.items()
            for (
                commit_messages,
                prerelease,
                expected_new_version,
                branch_name,
            ) in values  # type: ignore[attr-defined]
        ]
    ),
)
def test_version_next_greater_than_version_one_emoji(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    run_cli: RunCliFn,
    file_in_repo: str,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    [] if commit_messages is None else lazy_fixture(commit_messages),
                    prerelease,
                    prerelease_token,
                    expected_new_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_git_flow is currently 1.2.0-alpha.2
                    # The last full release version was 1.1.1, so it's had a minor
                    # prerelease
                    (
                        repo_w_git_flow_w_alpha_prereleases_n_emoji_commits.__name__,
                        "alpha",
                    ): [
                        *(
                            (commits, True, "1.2.0-alpha.2", "feat/feature-4")
                            for commits in (
                                None,
                                emoji_chore_commits.__name__,
                            )
                        ),
                        *(
                            (commits, False, "1.1.1", None)
                            for commits in (
                                None,
                                emoji_chore_commits.__name__,
                            )
                        ),
                    ],
                    # Latest version for repo_with_git_flow_and_release_channels is
                    # currently 1.1.0
                    (
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits.__name__,
                        "alpha",
                    ): [
                        *(
                            (commits, prerelease, "1.1.0", None)
                            for prerelease in (True, False)
                            for commits in (
                                None,
                                emoji_chore_commits.__name__,
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    expected_new_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ]
    ),
)
def test_version_next_greater_than_version_one_no_bump_emoji(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    run_cli: RunCliFn,
    file_in_repo: str,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # No commit has been made
    assert head_sha_before == head_after.hexsha
    assert len(tags_set_difference) == 0  # No tag created
    assert mocked_git_fetch.call_count == 0  # no git fetch called
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            pytest.param(
                lazy_fixture(repo_fixture_name),
                [] if commit_messages is None else lazy_fixture(commit_messages),
                prerelease,
                prerelease_token,
                expected_new_version,
                "main" if branch_name is None else branch_name,
                marks=pytest.mark.comprehensive,
            )
            for (repo_fixture_name, prerelease_token), values in {
                # Latest version for repo_with_git_flow is currently 1.2.0-alpha.2
                # The last full release version was 1.1.1, so it's had a minor
                # prerelease
                (
                    repo_w_git_flow_w_alpha_prereleases_n_scipy_commits.__name__,
                    "alpha",
                ): [
                    (scipy_patch_commits.__name__, False, "1.1.2", None),
                    (
                        scipy_patch_commits.__name__,
                        True,
                        "1.1.2-alpha.1",
                        None,
                    ),
                    (
                        scipy_minor_commits.__name__,
                        False,
                        "1.2.0",
                        None,
                    ),
                    (
                        scipy_minor_commits.__name__,
                        True,
                        "1.2.0-alpha.3",
                        "feat/feature-4",  # branch
                    ),
                    (scipy_major_commits.__name__, False, "2.0.0", None),
                    (
                        scipy_major_commits.__name__,
                        True,
                        "2.0.0-alpha.1",
                        None,
                    ),
                ],
                # Latest version for repo_with_git_flow_and_release_channels is
                # currently 1.1.0
                (
                    repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits.__name__,
                    "alpha",
                ): [
                    (scipy_patch_commits.__name__, False, "1.1.1", None),
                    (
                        scipy_patch_commits.__name__,
                        True,
                        "1.1.1-alpha.1",
                        None,
                    ),
                    (scipy_minor_commits.__name__, False, "1.2.0", None),
                    (
                        scipy_minor_commits.__name__,
                        True,
                        "1.2.0-alpha.1",
                        None,
                    ),
                    (scipy_major_commits.__name__, False, "2.0.0", None),
                    (
                        scipy_major_commits.__name__,
                        True,
                        "2.0.0-alpha.1",
                        None,
                    ),
                ],
            }.items()
            for (
                commit_messages,
                prerelease,
                expected_new_version,
                branch_name,
            ) in values  # type: ignore[attr-defined]
        ],
    ),
)
def test_version_next_greater_than_version_one_scipy(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    run_cli: RunCliFn,
    file_in_repo: str,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    [] if commit_messages is None else lazy_fixture(commit_messages),
                    prerelease,
                    prerelease_token,
                    expected_new_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_git_flow is currently 1.2.0-alpha.2
                    # The last full release version was 1.1.1, so it's had a minor
                    # prerelease
                    (
                        repo_w_git_flow_w_alpha_prereleases_n_scipy_commits.__name__,
                        "alpha",
                    ): [
                        *(
                            (commits, True, "1.2.0-alpha.2", "feat/feature-4")
                            for commits in (
                                None,
                                scipy_chore_commits.__name__,
                            )
                        ),
                        *(
                            (commits, False, "1.1.1", None)
                            for commits in (
                                None,
                                scipy_chore_commits.__name__,
                            )
                        ),
                    ],
                    # Latest version for repo_with_git_flow_and_release_channels is
                    # currently 1.1.0
                    (
                        repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits.__name__,
                        "alpha",
                    ): [
                        *(
                            (commits, prerelease, "1.1.0", None)
                            for prerelease in (True, False)
                            for commits in (
                                None,
                                scipy_chore_commits.__name__,
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    expected_new_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ]
    ),
)
def test_version_next_greater_than_version_one_no_bump_scipy(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    run_cli: RunCliFn,
    file_in_repo: str,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (no release actions should have occurred when no bump)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # No commit has been made
    assert head_sha_before == head_after.hexsha
    assert len(tags_set_difference) == 0  # No tag created
    assert mocked_git_fetch.call_count == 0  # no git fetch called
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


# ============================================================================= #
# Zero Dot version tests (ex. 0.x.y versions)
# ============================================================================= #


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "major_on_zero",
            "allow_zero_version",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            (
                # Latest version for repo_with_no_tags is currently 0.0.0 (default)
                # It's biggest change type is minor, so the next version should be 0.1.0
                # Given the major_on_zero is False and the version is starting at 0.0.0,
                # the major level commits are limited to only causing a minor level bump
                lazy_fixture(repo_w_no_tags_conventional_commits.__name__),
                lazy_fixture(conventional_major_commits.__name__),
                False,
                "rc",
                False,
                True,
                "0.1.0",
                "main",
            ),
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    commit_messages,
                    prerelease,
                    "rc" if prerelease_token is None else prerelease_token,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_no_tags is currently 0.0.0 (default)
                    # It's biggest change type is minor, so the next version should be 0.1.0
                    (
                        repo_w_no_tags_conventional_commits.__name__,
                        None,
                    ): [
                        *(
                            # when prerelease is False, & major_on_zero is False &
                            # allow_zero_version is True, the version should be
                            # 0.1.0, with the given commits
                            (commits, False, False, True, "0.1.0", None)
                            for commits in (
                                # Even when this test does not change anything, the base modification
                                # will be a minor change and thus the version will be bumped to 0.1.0
                                None,
                                # Non version bumping commits are absorbed into the previously detected minor bump
                                lazy_fixture(conventional_chore_commits.__name__),
                                # Patch commits are absorbed into the previously detected minor bump
                                lazy_fixture(conventional_patch_commits.__name__),
                                # Minor level commits are absorbed into the previously detected minor bump
                                lazy_fixture(conventional_minor_commits.__name__),
                                # Given the major_on_zero is False and the version is starting at 0.0.0,
                                # the major level commits are limited to only causing a minor level bump
                                # lazy_fixture(conventional_major_commits.__name__), # used as default
                            )
                        ),
                        # when prerelease is False, & major_on_zero is False, & allow_zero_version is True,
                        # the version should only be minor bumped when provided major commits because
                        # of the major_on_zero value
                        (
                            lazy_fixture(conventional_major_commits.__name__),
                            False,
                            False,
                            True,
                            "0.1.0",
                            None,
                        ),
                        # when prerelease is False, & major_on_zero is True & allow_zero_version is True,
                        # the version should be major bumped when provided major commits because
                        # of the major_on_zero value
                        (
                            lazy_fixture(conventional_major_commits.__name__),
                            False,
                            True,
                            True,
                            "1.0.0",
                            None,
                        ),
                        *(
                            # when prerelease is False, & allow_zero_version is False, the version should be
                            # 1.0.0, across the board because 0 is not a valid major version.
                            # major_on_zero is ignored as it is not relevant but tested for completeness
                            (commits, False, major_on_zero, False, "1.0.0", None)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(conventional_chore_commits.__name__),
                                lazy_fixture(conventional_patch_commits.__name__),
                                lazy_fixture(conventional_minor_commits.__name__),
                                lazy_fixture(conventional_major_commits.__name__),
                            )
                        ),
                    ],
                    # Latest version for repo_with_single_branch is currently 0.1.1
                    # Note repo_with_single_branch isn't modelled with prereleases
                    (
                        repo_w_trunk_only_conventional_commits.__name__,
                        None,
                    ): [
                        *(
                            # when prerelease must be False, and allow_zero_version is True,
                            # the version is patch bumped because of the patch level commits
                            # regardless of the major_on_zero value
                            (
                                lazy_fixture(conventional_patch_commits.__name__),
                                False,
                                major_on_zero,
                                True,
                                "0.1.2",
                                None,
                            )
                            for major_on_zero in (True, False)
                        ),
                        *(
                            # when prerelease must be False, and allow_zero_version is True,
                            # the version is minor bumped because of the major_on_zero value=False
                            (commits, False, False, True, "0.2.0", None)
                            for commits in (
                                lazy_fixture(conventional_minor_commits.__name__),
                                lazy_fixture(conventional_major_commits.__name__),
                            )
                        ),
                        # when prerelease must be False, and allow_zero_version is True,
                        # but the major_on_zero is True, then when a major level commit is given,
                        # the version should be bumped to the next major version
                        (
                            lazy_fixture(conventional_major_commits.__name__),
                            False,
                            True,
                            True,
                            "1.0.0",
                            None,
                        ),
                        *(
                            # when prerelease must be False, & allow_zero_version is False, the version should be
                            # 1.0.0, with any change regardless of major_on_zero
                            (commits, False, major_on_zero, False, "1.0.0", None)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(conventional_chore_commits.__name__),
                                lazy_fixture(conventional_patch_commits.__name__),
                                lazy_fixture(conventional_minor_commits.__name__),
                                lazy_fixture(conventional_major_commits.__name__),
                            )
                        ),
                    ],
                    # Latest version for repo_with_single_branch_and_prereleases is
                    # currently 0.2.0
                    (
                        repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                        None,
                    ): [
                        # when allow_zero_version is True,
                        # prerelease is False, & major_on_zero is False, the version should be
                        # patch bumped as a prerelease version, when given patch level commits
                        (
                            lazy_fixture(conventional_patch_commits.__name__),
                            True,
                            False,
                            True,
                            "0.2.1-rc.1",
                            None,
                        ),
                        # when allow_zero_version is True,
                        # prerelease is False, & major_on_zero is False, the version should be
                        # patch bumped, when given patch level commits
                        (
                            lazy_fixture(conventional_patch_commits.__name__),
                            False,
                            False,
                            True,
                            "0.2.1",
                            None,
                        ),
                        *(
                            # when allow_zero_version is True,
                            # prerelease is True, & major_on_zero is False, the version should be
                            # minor bumped as a prerelease version, when given commits of a minor or major level
                            (commits, True, False, True, "0.3.0-rc.1", None)
                            for commits in (
                                lazy_fixture(conventional_minor_commits.__name__),
                                lazy_fixture(conventional_major_commits.__name__),
                            )
                        ),
                        *(
                            # when allow_zero_version is True, prerelease is True, & major_on_zero
                            # is False, the version should be minor bumped, when given commits of a
                            # minor or major level because major_on_zero = False
                            (commits, False, False, True, "0.3.0", None)
                            for commits in (
                                lazy_fixture(conventional_minor_commits.__name__),
                                lazy_fixture(conventional_major_commits.__name__),
                            )
                        ),
                        # when prerelease is True, & major_on_zero is True, and allow_zero_version
                        # is True, the version should be bumped to 1.0.0 as a prerelease version, when
                        # given major level commits
                        (
                            lazy_fixture(conventional_major_commits.__name__),
                            True,
                            True,
                            True,
                            "1.0.0-rc.1",
                            None,
                        ),
                        # when prerelease is False, & major_on_zero is True, and allow_zero_version
                        # is True, the version should be bumped to 1.0.0, when given major level commits
                        (
                            lazy_fixture(conventional_major_commits.__name__),
                            False,
                            True,
                            True,
                            "1.0.0",
                            None,
                        ),
                        *(
                            # when prerelease is True, & allow_zero_version is False, the version should be
                            # bumped to 1.0.0 as a prerelease version, when given any/none commits
                            # because 0.x is no longer a valid version regardless of the major_on_zero value
                            (commits, True, major_on_zero, False, "1.0.0-rc.1", None)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(conventional_chore_commits.__name__),
                                lazy_fixture(conventional_patch_commits.__name__),
                                lazy_fixture(conventional_minor_commits.__name__),
                                lazy_fixture(conventional_major_commits.__name__),
                            )
                        ),
                        *(
                            # when prerelease is True, & allow_zero_version is False, the version should be
                            # bumped to 1.0.0, when given any/none commits
                            # because 0.x is no longer a valid version regardless of the major_on_zero value
                            (commits, False, major_on_zero, False, "1.0.0", None)
                            for major_on_zero in (True, False)
                            for commits in (
                                lazy_fixture(conventional_patch_commits.__name__),
                                lazy_fixture(conventional_minor_commits.__name__),
                                lazy_fixture(conventional_major_commits.__name__),
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ],
    ),
)
def test_version_next_w_zero_dot_versions_conventional(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    major_on_zero: bool,
    allow_zero_version: bool,
    run_cli: RunCliFn,
    file_in_repo: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: update pyproject.toml with the necessary settings
    update_pyproject_toml(
        "tool.semantic_release.allow_zero_version", allow_zero_version
    )
    update_pyproject_toml("tool.semantic_release.major_on_zero", major_on_zero)

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "major_on_zero",
            "allow_zero_version",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    commit_messages,
                    prerelease,
                    "rc" if prerelease_token is None else prerelease_token,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_single_branch is currently 0.1.1
                    # Note repo_with_single_branch isn't modelled with prereleases
                    (
                        repo_w_trunk_only_conventional_commits.__name__,
                        None,
                    ): [
                        *(
                            # when prerelease must be False, and allow_zero_version is True,
                            # the version is not bumped because of non valuable changes regardless
                            # of the major_on_zero value
                            (commits, False, major_on_zero, True, "0.1.1", None)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(conventional_chore_commits.__name__),
                            )
                        ),
                    ],
                    # Latest version for repo_with_single_branch_and_prereleases is
                    # currently 0.2.0
                    (
                        repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                        None,
                    ): [
                        *(
                            # when allow_zero_version is True, the version is not bumped
                            # regardless of prerelease and major_on_zero values when given
                            # non valuable changes
                            (commits, prerelease, major_on_zero, True, "0.2.0", None)
                            for prerelease in (True, False)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(conventional_chore_commits.__name__),
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ],
    ),
)
def test_version_next_w_zero_dot_versions_no_bump_conventional(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    major_on_zero: bool,
    allow_zero_version: bool,
    run_cli: RunCliFn,
    file_in_repo: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: update pyproject.toml with the necessary settings
    update_pyproject_toml(
        "tool.semantic_release.allow_zero_version", allow_zero_version
    )
    update_pyproject_toml("tool.semantic_release.major_on_zero", major_on_zero)

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (no release actions should have occurred when no bump)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # No commit has been made
    assert head_sha_before == head_after.hexsha
    assert len(tags_set_difference) == 0  # No tag created
    assert mocked_git_fetch.call_count == 0  # no git fetch called
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "major_on_zero",
            "allow_zero_version",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            pytest.param(
                lazy_fixture(repo_fixture_name),
                commit_messages,
                prerelease,
                "rc" if prerelease_token is None else prerelease_token,
                major_on_zero,
                allow_zero_version,
                next_release_version,
                "main" if branch_name is None else branch_name,
                marks=pytest.mark.comprehensive,
            )
            for (repo_fixture_name, prerelease_token), values in {
                # Latest version for repo_with_no_tags is currently 0.0.0 (default)
                # It's biggest change type is minor, so the next version should be 0.1.0
                (
                    repo_w_no_tags_emoji_commits.__name__,
                    None,
                ): [
                    *(
                        # when prerelease is False, & major_on_zero is False &
                        # allow_zero_version is True, the version should be
                        # 0.1.0, with the given commits
                        (commits, False, False, True, "0.1.0", None)
                        for commits in (
                            # Even when this test does not change anything, the base modification
                            # will be a minor change and thus the version will be bumped to 0.1.0
                            None,
                            # Non version bumping commits are absorbed into the previously detected minor bump
                            lazy_fixture(emoji_chore_commits.__name__),
                            # Patch commits are absorbed into the previously detected minor bump
                            lazy_fixture(emoji_patch_commits.__name__),
                            # Minor level commits are absorbed into the previously detected minor bump
                            lazy_fixture(emoji_minor_commits.__name__),
                            # Given the major_on_zero is False and the version is starting at 0.0.0,
                            # the major level commits are limited to only causing a minor level bump
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                    # when prerelease is False, & major_on_zero is False, & allow_zero_version is True,
                    # the version should only be minor bumped when provided major commits because
                    # of the major_on_zero value
                    (
                        lazy_fixture(emoji_major_commits.__name__),
                        False,
                        False,
                        True,
                        "0.1.0",
                        None,
                    ),
                    # when prerelease is False, & major_on_zero is True & allow_zero_version is True,
                    # the version should be major bumped when provided major commits because
                    # of the major_on_zero value
                    (
                        lazy_fixture(emoji_major_commits.__name__),
                        False,
                        True,
                        True,
                        "1.0.0",
                        None,
                    ),
                    *(
                        # when prerelease is False, & allow_zero_version is False, the version should be
                        # 1.0.0, across the board because 0 is not a valid major version.
                        # major_on_zero is ignored as it is not relevant but tested for completeness
                        (commits, False, major_on_zero, False, "1.0.0", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            None,
                            lazy_fixture(emoji_chore_commits.__name__),
                            lazy_fixture(emoji_patch_commits.__name__),
                            lazy_fixture(emoji_minor_commits.__name__),
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                ],
                # Latest version for repo_with_single_branch is currently 0.1.1
                # Note repo_with_single_branch isn't modelled with prereleases
                (
                    repo_w_trunk_only_emoji_commits.__name__,
                    None,
                ): [
                    *(
                        # when prerelease must be False, and allow_zero_version is True,
                        # the version is patch bumped because of the patch level commits
                        # regardless of the major_on_zero value
                        (
                            lazy_fixture(emoji_patch_commits.__name__),
                            False,
                            major_on_zero,
                            True,
                            "0.1.2",
                            None,
                        )
                        for major_on_zero in (True, False)
                    ),
                    *(
                        # when prerelease must be False, and allow_zero_version is True,
                        # the version is minor bumped because of the major_on_zero value=False
                        (commits, False, False, True, "0.2.0", None)
                        for commits in (
                            lazy_fixture(emoji_minor_commits.__name__),
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                    # when prerelease must be False, and allow_zero_version is True,
                    # but the major_on_zero is True, then when a major level commit is given,
                    # the version should be bumped to the next major version
                    (
                        lazy_fixture(emoji_major_commits.__name__),
                        False,
                        True,
                        True,
                        "1.0.0",
                        None,
                    ),
                    *(
                        # when prerelease must be False, & allow_zero_version is False, the version should be
                        # 1.0.0, with any change regardless of major_on_zero
                        (commits, False, major_on_zero, False, "1.0.0", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            None,
                            lazy_fixture(emoji_chore_commits.__name__),
                            lazy_fixture(emoji_patch_commits.__name__),
                            lazy_fixture(emoji_minor_commits.__name__),
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                ],
                # Latest version for repo_with_single_branch_and_prereleases is
                # currently 0.2.0
                (
                    repo_w_trunk_only_n_prereleases_emoji_commits.__name__,
                    None,
                ): [
                    # when allow_zero_version is True,
                    # prerelease is False, & major_on_zero is False, the version should be
                    # patch bumped as a prerelease version, when given patch level commits
                    (
                        lazy_fixture(emoji_patch_commits.__name__),
                        True,
                        False,
                        True,
                        "0.2.1-rc.1",
                        None,
                    ),
                    # when allow_zero_version is True,
                    # prerelease is False, & major_on_zero is False, the version should be
                    # patch bumped, when given patch level commits
                    (
                        lazy_fixture(emoji_patch_commits.__name__),
                        False,
                        False,
                        True,
                        "0.2.1",
                        None,
                    ),
                    *(
                        # when allow_zero_version is True,
                        # prerelease is True, & major_on_zero is False, the version should be
                        # minor bumped as a prerelease version, when given commits of a minor or major level
                        (commits, True, False, True, "0.3.0-rc.1", None)
                        for commits in (
                            lazy_fixture(emoji_minor_commits.__name__),
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                    *(
                        # when allow_zero_version is True, prerelease is True, & major_on_zero
                        # is False, the version should be minor bumped, when given commits of a
                        # minor or major level because major_on_zero = False
                        (commits, False, False, True, "0.3.0", None)
                        for commits in (
                            lazy_fixture(emoji_minor_commits.__name__),
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                    # when prerelease is True, & major_on_zero is True, and allow_zero_version
                    # is True, the version should be bumped to 1.0.0 as a prerelease version, when
                    # given major level commits
                    (
                        lazy_fixture(emoji_major_commits.__name__),
                        True,
                        True,
                        True,
                        "1.0.0-rc.1",
                        None,
                    ),
                    # when prerelease is False, & major_on_zero is True, and allow_zero_version
                    # is True, the version should be bumped to 1.0.0, when given major level commits
                    (
                        lazy_fixture(emoji_major_commits.__name__),
                        False,
                        True,
                        True,
                        "1.0.0",
                        None,
                    ),
                    *(
                        # when prerelease is True, & allow_zero_version is False, the version should be
                        # bumped to 1.0.0 as a prerelease version, when given any/none commits
                        # because 0.x is no longer a valid version regardless of the major_on_zero value
                        (commits, True, major_on_zero, False, "1.0.0-rc.1", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            None,
                            lazy_fixture(emoji_chore_commits.__name__),
                            lazy_fixture(emoji_patch_commits.__name__),
                            lazy_fixture(emoji_minor_commits.__name__),
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                    *(
                        # when prerelease is True, & allow_zero_version is False, the version should be
                        # bumped to 1.0.0, when given any/none commits
                        # because 0.x is no longer a valid version regardless of the major_on_zero value
                        (commits, False, major_on_zero, False, "1.0.0", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            lazy_fixture(emoji_patch_commits.__name__),
                            lazy_fixture(emoji_minor_commits.__name__),
                            lazy_fixture(emoji_major_commits.__name__),
                        )
                    ),
                ],
            }.items()
            for (
                commit_messages,
                prerelease,
                major_on_zero,
                allow_zero_version,
                next_release_version,
                branch_name,
            ) in values  # type: ignore[attr-defined]
        ],
    ),
)
def test_version_next_w_zero_dot_versions_emoji(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    major_on_zero: bool,
    allow_zero_version: bool,
    run_cli: RunCliFn,
    file_in_repo: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: update pyproject.toml with the necessary settings
    update_pyproject_toml(
        "tool.semantic_release.allow_zero_version", allow_zero_version
    )
    update_pyproject_toml("tool.semantic_release.major_on_zero", major_on_zero)

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "major_on_zero",
            "allow_zero_version",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    commit_messages,
                    prerelease,
                    "rc" if prerelease_token is None else prerelease_token,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_single_branch is currently 0.1.1
                    # Note repo_with_single_branch isn't modelled with prereleases
                    (
                        repo_w_trunk_only_emoji_commits.__name__,
                        None,
                    ): [
                        *(
                            # when prerelease must be False, and allow_zero_version is True,
                            # the version is not bumped because of non valuable changes regardless
                            # of the major_on_zero value
                            (commits, False, major_on_zero, True, "0.1.1", None)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(emoji_chore_commits.__name__),
                            )
                        ),
                    ],
                    # Latest version for repo_with_single_branch_and_prereleases is
                    # currently 0.2.0
                    (
                        repo_w_trunk_only_n_prereleases_emoji_commits.__name__,
                        None,
                    ): [
                        *(
                            # when allow_zero_version is True, the version is not bumped
                            # regardless of prerelease and major_on_zero values when given
                            # non valuable changes
                            (commits, prerelease, major_on_zero, True, "0.2.0", None)
                            for prerelease in (True, False)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(emoji_chore_commits.__name__),
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ],
    ),
)
def test_version_next_w_zero_dot_versions_no_bump_emoji(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    major_on_zero: bool,
    allow_zero_version: bool,
    run_cli: RunCliFn,
    file_in_repo: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: update pyproject.toml with the necessary settings
    update_pyproject_toml(
        "tool.semantic_release.allow_zero_version", allow_zero_version
    )
    update_pyproject_toml("tool.semantic_release.major_on_zero", major_on_zero)

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (no release actions should have occurred when no bump)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # No commit has been made
    assert head_sha_before == head_after.hexsha
    assert len(tags_set_difference) == 0  # No tag created
    assert mocked_git_fetch.call_count == 0  # no git fetch called
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "major_on_zero",
            "allow_zero_version",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            pytest.param(
                lazy_fixture(repo_fixture_name),
                commit_messages,
                prerelease,
                "rc" if prerelease_token is None else prerelease_token,
                major_on_zero,
                allow_zero_version,
                next_release_version,
                "main" if branch_name is None else branch_name,
                marks=pytest.mark.comprehensive,
            )
            for (repo_fixture_name, prerelease_token), values in {
                # Latest version for repo_with_no_tags is currently 0.0.0 (default)
                # It's biggest change type is minor, so the next version should be 0.1.0
                (
                    repo_w_no_tags_scipy_commits.__name__,
                    None,
                ): [
                    *(
                        # when prerelease is False, & major_on_zero is False &
                        # allow_zero_version is True, the version should be
                        # 0.1.0, with the given commits
                        (commits, False, False, True, "0.1.0", None)
                        for commits in (
                            # Even when this test does not change anything, the base modification
                            # will be a minor change and thus the version will be bumped to 0.1.0
                            None,
                            # Non version bumping commits are absorbed into the previously detected minor bump
                            lazy_fixture(scipy_chore_commits.__name__),
                            # Patch commits are absorbed into the previously detected minor bump
                            lazy_fixture(scipy_patch_commits.__name__),
                            # Minor level commits are absorbed into the previously detected minor bump
                            lazy_fixture(scipy_minor_commits.__name__),
                            # Given the major_on_zero is False and the version is starting at 0.0.0,
                            # the major level commits are limited to only causing a minor level bump
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                    # when prerelease is False, & major_on_zero is False, & allow_zero_version is True,
                    # the version should only be minor bumped when provided major commits because
                    # of the major_on_zero value
                    (
                        lazy_fixture(scipy_major_commits.__name__),
                        False,
                        False,
                        True,
                        "0.1.0",
                        None,
                    ),
                    # when prerelease is False, & major_on_zero is True & allow_zero_version is True,
                    # the version should be major bumped when provided major commits because
                    # of the major_on_zero value
                    (
                        lazy_fixture(scipy_major_commits.__name__),
                        False,
                        True,
                        True,
                        "1.0.0",
                        None,
                    ),
                    *(
                        # when prerelease is False, & allow_zero_version is False, the version should be
                        # 1.0.0, across the board because 0 is not a valid major version.
                        # major_on_zero is ignored as it is not relevant but tested for completeness
                        (commits, False, major_on_zero, False, "1.0.0", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            None,
                            lazy_fixture(scipy_chore_commits.__name__),
                            lazy_fixture(scipy_patch_commits.__name__),
                            lazy_fixture(scipy_minor_commits.__name__),
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                ],
                # Latest version for repo_with_single_branch is currently 0.1.1
                # Note repo_with_single_branch isn't modelled with prereleases
                (
                    repo_w_trunk_only_scipy_commits.__name__,
                    None,
                ): [
                    *(
                        # when prerelease must be False, and allow_zero_version is True,
                        # the version is patch bumped because of the patch level commits
                        # regardless of the major_on_zero value
                        (
                            lazy_fixture(scipy_patch_commits.__name__),
                            False,
                            major_on_zero,
                            True,
                            "0.1.2",
                            None,
                        )
                        for major_on_zero in (True, False)
                    ),
                    *(
                        # when prerelease must be False, and allow_zero_version is True,
                        # the version is minor bumped because of the major_on_zero value=False
                        (commits, False, False, True, "0.2.0", None)
                        for commits in (
                            lazy_fixture(scipy_minor_commits.__name__),
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                    # when prerelease must be False, and allow_zero_version is True,
                    # but the major_on_zero is True, then when a major level commit is given,
                    # the version should be bumped to the next major version
                    (
                        lazy_fixture(scipy_major_commits.__name__),
                        False,
                        True,
                        True,
                        "1.0.0",
                        None,
                    ),
                    *(
                        # when prerelease must be False, & allow_zero_version is False, the version should be
                        # 1.0.0, with any change regardless of major_on_zero
                        (commits, False, major_on_zero, False, "1.0.0", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            None,
                            lazy_fixture(scipy_chore_commits.__name__),
                            lazy_fixture(scipy_patch_commits.__name__),
                            lazy_fixture(scipy_minor_commits.__name__),
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                ],
                # Latest version for repo_with_single_branch_and_prereleases is
                # currently 0.2.0
                (
                    repo_w_trunk_only_n_prereleases_scipy_commits.__name__,
                    None,
                ): [
                    # when allow_zero_version is True,
                    # prerelease is False, & major_on_zero is False, the version should be
                    # patch bumped as a prerelease version, when given patch level commits
                    (
                        lazy_fixture(scipy_patch_commits.__name__),
                        True,
                        False,
                        True,
                        "0.2.1-rc.1",
                        None,
                    ),
                    # when allow_zero_version is True,
                    # prerelease is False, & major_on_zero is False, the version should be
                    # patch bumped, when given patch level commits
                    (
                        lazy_fixture(scipy_patch_commits.__name__),
                        False,
                        False,
                        True,
                        "0.2.1",
                        None,
                    ),
                    *(
                        # when allow_zero_version is True,
                        # prerelease is True, & major_on_zero is False, the version should be
                        # minor bumped as a prerelease version, when given commits of a minor or major level
                        (commits, True, False, True, "0.3.0-rc.1", None)
                        for commits in (
                            lazy_fixture(scipy_minor_commits.__name__),
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                    *(
                        # when allow_zero_version is True, prerelease is True, & major_on_zero
                        # is False, the version should be minor bumped, when given commits of a
                        # minor or major level because major_on_zero = False
                        (commits, False, False, True, "0.3.0", None)
                        for commits in (
                            lazy_fixture(scipy_minor_commits.__name__),
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                    # when prerelease is True, & major_on_zero is True, and allow_zero_version
                    # is True, the version should be bumped to 1.0.0 as a prerelease version, when
                    # given major level commits
                    (
                        lazy_fixture(scipy_major_commits.__name__),
                        True,
                        True,
                        True,
                        "1.0.0-rc.1",
                        None,
                    ),
                    # when prerelease is False, & major_on_zero is True, and allow_zero_version
                    # is True, the version should be bumped to 1.0.0, when given major level commits
                    (
                        lazy_fixture(scipy_major_commits.__name__),
                        False,
                        True,
                        True,
                        "1.0.0",
                        None,
                    ),
                    *(
                        # when prerelease is True, & allow_zero_version is False, the version should be
                        # bumped to 1.0.0 as a prerelease version, when given any/none commits
                        # because 0.x is no longer a valid version regardless of the major_on_zero value
                        (commits, True, major_on_zero, False, "1.0.0-rc.1", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            None,
                            lazy_fixture(scipy_chore_commits.__name__),
                            lazy_fixture(scipy_patch_commits.__name__),
                            lazy_fixture(scipy_minor_commits.__name__),
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                    *(
                        # when prerelease is True, & allow_zero_version is False, the version should be
                        # bumped to 1.0.0, when given any/none commits
                        # because 0.x is no longer a valid version regardless of the major_on_zero value
                        (commits, False, major_on_zero, False, "1.0.0", None)
                        for major_on_zero in (True, False)
                        for commits in (
                            lazy_fixture(scipy_patch_commits.__name__),
                            lazy_fixture(scipy_minor_commits.__name__),
                            lazy_fixture(scipy_major_commits.__name__),
                        )
                    ),
                ],
            }.items()
            for (
                commit_messages,
                prerelease,
                major_on_zero,
                allow_zero_version,
                next_release_version,
                branch_name,
            ) in values  # type: ignore[attr-defined]
        ],
    ),
)
def test_version_next_w_zero_dot_versions_scipy(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    major_on_zero: bool,
    allow_zero_version: bool,
    run_cli: RunCliFn,
    file_in_repo: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: update pyproject.toml with the necessary settings
    update_pyproject_toml(
        "tool.semantic_release.allow_zero_version", allow_zero_version
    )
    update_pyproject_toml("tool.semantic_release.major_on_zero", major_on_zero)

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "repo_result",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "major_on_zero",
            "allow_zero_version",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            *[
                pytest.param(
                    lazy_fixture(repo_fixture_name),
                    commit_messages,
                    prerelease,
                    "rc" if prerelease_token is None else prerelease_token,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    "main" if branch_name is None else branch_name,
                    marks=pytest.mark.comprehensive,
                )
                for (repo_fixture_name, prerelease_token), values in {
                    # Latest version for repo_with_single_branch is currently 0.1.1
                    # Note repo_with_single_branch isn't modelled with prereleases
                    (
                        repo_w_trunk_only_scipy_commits.__name__,
                        None,
                    ): [
                        *(
                            # when prerelease must be False, and allow_zero_version is True,
                            # the version is not bumped because of non valuable changes regardless
                            # of the major_on_zero value
                            (commits, False, major_on_zero, True, "0.1.1", None)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(scipy_chore_commits.__name__),
                            )
                        ),
                    ],
                    # Latest version for repo_with_single_branch_and_prereleases is
                    # currently 0.2.0
                    (
                        repo_w_trunk_only_n_prereleases_scipy_commits.__name__,
                        None,
                    ): [
                        *(
                            # when allow_zero_version is True, the version is not bumped
                            # regardless of prerelease and major_on_zero values when given
                            # non valuable changes
                            (commits, prerelease, major_on_zero, True, "0.2.0", None)
                            for prerelease in (True, False)
                            for major_on_zero in (True, False)
                            for commits in (
                                None,
                                lazy_fixture(scipy_chore_commits.__name__),
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    prerelease,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    branch_name,
                ) in values  # type: ignore[attr-defined]
            ],
        ],
    ),
)
def test_version_next_w_zero_dot_versions_no_bump_scipy(
    repo_result: BuiltRepoResult,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    major_on_zero: bool,
    allow_zero_version: bool,
    run_cli: RunCliFn,
    file_in_repo: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: update pyproject.toml with the necessary settings
    update_pyproject_toml(
        "tool.semantic_release.allow_zero_version", allow_zero_version
    )
    update_pyproject_toml("tool.semantic_release.major_on_zero", major_on_zero)

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (no release actions should have occurred when no bump)
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # No commit has been made
    assert head_sha_before == head_after.hexsha
    assert len(tags_set_difference) == 0  # No tag created
    assert mocked_git_fetch.call_count == 0  # no git fetch called
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    str.join(
        " ,",
        [
            "repo_result",
            "commit_parser",
            "commit_messages",
            "prerelease",
            "prerelease_token",
            "major_on_zero",
            "allow_zero_version",
            "next_release_version",
            "branch_name",
        ],
    ),
    xdist_sort_hack(
        [
            (
                # Latest version for repo_w_initial_commit is currently 0.0.0
                # with no changes made it should be 0.0.0
                lazy_fixture(repo_w_initial_commit.__name__),
                ConventionalCommitParser.__name__.replace("CommitParser", "").lower(),
                None,
                False,
                "rc",
                False,
                True,
                "0.0.0",
                "main",
            ),
            *[
                pytest.param(
                    lazy_fixture(repo_w_initial_commit.__name__),
                    str.replace(parser_class_name, "CommitParser", "").lower(),
                    commit_messages,
                    prerelease,
                    prerelease_token,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                    "main",
                    marks=pytest.mark.comprehensive,
                )
                for prerelease_token, values in {
                    # Latest version for repo_with_no_tags is currently 0.0.0 (default)
                    # It's biggest change type is minor, so the next version should be 0.1.0
                    "rc": [
                        *(
                            # when prerelease is False, major_on_zero is True & False, & allow_zero_version is True
                            # the version should be 0.0.0, when no distintive changes have been made since the
                            # start of the project
                            (commits, parser, prerelease, major_on_zero, True, "0.0.0")
                            for prerelease in (True, False)
                            for major_on_zero in (True, False)
                            for commits, parser in (
                                # No commits added, so base is just initial commit at 0.0.0
                                (None, ConventionalCommitParser.__name__),
                                # Chore like commits also don't trigger a version bump so it stays 0.0.0
                                (
                                    lazy_fixture(conventional_chore_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_chore_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_chore_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                            )
                        ),
                        *(
                            (commits, parser, True, major_on_zero, True, "0.0.1-rc.1")
                            for major_on_zero in (True, False)
                            for commits, parser in (
                                # when prerelease is True & allow_zero_version is True, the version should be
                                # a patch bump as a prerelease version, because of the patch level commits
                                # major_on_zero is irrelevant here as we are only applying patch commits
                                (
                                    lazy_fixture(conventional_patch_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_patch_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_patch_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                            )
                        ),
                        *(
                            (commits, parser, False, major_on_zero, True, "0.0.1")
                            for major_on_zero in (True, False)
                            for commits, parser in (
                                # when prerelease is False, & allow_zero_version is True, the version should be
                                # a patch bump because of the patch commits added
                                # major_on_zero is irrelevant here as we are only applying patch commits
                                (
                                    lazy_fixture(conventional_patch_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_patch_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_patch_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                            )
                        ),
                        *(
                            (commits, parser, True, False, True, "0.1.0-rc.1")
                            for commits, parser in (
                                # when prerelease is False, & major_on_zero is False, the version should be
                                # a minor bump because of the minor commits added
                                (
                                    lazy_fixture(conventional_minor_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_minor_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_minor_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                # Given the major_on_zero is False and the version is starting at 0.0.0,
                                # the major level commits are limited to only causing a minor level bump
                                (
                                    lazy_fixture(conventional_major_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_major_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_major_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                            )
                        ),
                        *(
                            (commits, parser, False, False, True, "0.1.0")
                            for commits, parser in (
                                # when prerelease is False,
                                # major_on_zero is False, & allow_zero_version is True
                                # the version should be a minor bump of 0.0.0
                                # because of the minor commits added and zero version is allowed
                                (
                                    lazy_fixture(conventional_minor_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_minor_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_minor_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                # Given the major_on_zero is False and the version is starting at 0.0.0,
                                # the major level commits are limited to only causing a minor level bump
                                (
                                    lazy_fixture(conventional_major_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_major_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_major_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                            )
                        ),
                        *(
                            # when prerelease is True, & allow_zero_version is False, the version should be
                            # a prerelease version 1.0.0-rc.1, across the board when any valuable change
                            # is made because of the allow_zero_version is False, major_on_zero is ignored
                            # when allow_zero_version is False (but we still test it)
                            (commits, parser, True, major_on_zero, False, "1.0.0-rc.1")
                            for major_on_zero in (True, False)
                            for commits, parser in (
                                # parser doesn't matter here as long as it detects a NO_RELEASE on Initial Commit
                                (None, ConventionalCommitParser.__name__),
                                (
                                    lazy_fixture(conventional_chore_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(conventional_patch_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(conventional_minor_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(conventional_major_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_chore_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_patch_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_minor_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_major_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_chore_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_patch_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_minor_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_major_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                            )
                        ),
                        *(
                            # when prerelease is True, & allow_zero_version is False, the version should be
                            # 1.0.0, across the board when any valuable change
                            # is made because of the allow_zero_version is False. major_on_zero is ignored
                            # when allow_zero_version is False (but we still test it)
                            (commits, parser, False, major_on_zero, False, "1.0.0")
                            for major_on_zero in (True, False)
                            for commits, parser in (
                                (None, ConventionalCommitParser.__name__),
                                (
                                    lazy_fixture(conventional_chore_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(conventional_patch_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(conventional_minor_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(conventional_major_commits.__name__),
                                    ConventionalCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_chore_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_patch_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_minor_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(emoji_major_commits.__name__),
                                    EmojiCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_chore_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_patch_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_minor_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                                (
                                    lazy_fixture(scipy_major_commits.__name__),
                                    ScipyCommitParser.__name__,
                                ),
                            )
                        ),
                    ],
                }.items()
                for (
                    commit_messages,
                    parser_class_name,
                    prerelease,
                    major_on_zero,
                    allow_zero_version,
                    next_release_version,
                ) in values  # type: ignore[attr-defined]
            ],
        ],
    ),
)
def test_version_next_w_zero_dot_versions_minimums(
    repo_result: BuiltRepoResult,
    commit_parser: str,
    commit_messages: list[str],
    prerelease: bool,
    prerelease_token: str,
    next_release_version: str,
    branch_name: str,
    major_on_zero: bool,
    allow_zero_version: bool,
    run_cli: RunCliFn,
    file_in_repo: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    stable_now_date: GetStableDateNowFn,
):
    repo = repo_result["repo"]

    # setup: select the branch we desire for the next bump
    if repo.active_branch.name != branch_name:
        repo.heads[branch_name].checkout()

    # setup: update pyproject.toml with the necessary settings
    update_pyproject_toml("tool.semantic_release.commit_parser", commit_parser)
    update_pyproject_toml(
        "tool.semantic_release.allow_zero_version", allow_zero_version
    )
    update_pyproject_toml("tool.semantic_release.major_on_zero", major_on_zero)

    # setup: apply commits to the repo
    stable_now_datetime = stable_now_date()
    commit_timestamp_gen = (
        (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
        for i in count(step=1)
    )
    for commit_message in commit_messages or []:
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commit_message, a=True, date=next(commit_timestamp_gen))
        # Fake an automated push to remote by updating the remote tracking branch
        repo.git.update_ref(
            f"refs/remotes/origin/{repo.active_branch.name}",
            repo.head.commit.hexsha,
        )

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Derive the cli arguments based on parameter input
    prerelease_args = list(
        filter(
            None,
            [
                "--as-prerelease" if prerelease else "",
                *(["--prerelease-token", prerelease_token] if prerelease else []),
            ],
        )
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *prerelease_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made (regardless of precommit)
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    assert len(tags_set_difference) == 1  # A tag has been created
    assert f"v{next_release_version}" in tags_set_difference

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
    assert post_mocker.call_count == 1  # vcs release creation occurred



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import pytest
import tomlkit

from semantic_release.cli.config import RawConfig

from tests.const import GENERATE_CONFIG_SUBCMD, MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.repos import repo_w_no_tags_conventional_commits
from tests.util import assert_successful_exit_code

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import ExProjectDir


@pytest.fixture
def raw_config_dict() -> dict[str, Any]:
    return RawConfig().model_dump(mode="json", exclude_none=True)


@pytest.mark.parametrize("args", [(), ("--format", "toml"), ("--format", "TOML")])
@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_generate_config_toml(
    run_cli: RunCliFn,
    args: tuple[str],
    raw_config_dict: dict[str, Any],
    example_project_dir: ExProjectDir,
):
    # Setup: Generate the expected configuration as a TOML string
    expected_config_as_str = tomlkit.dumps(
        {"semantic_release": raw_config_dict}
    ).strip()

    # Act: Print the generated configuration to stdout
    cli_cmd = [MAIN_PROG_NAME, GENERATE_CONFIG_SUBCMD, *args]
    result = run_cli(cli_cmd[1:])

    # Evaluate: Check that the command ran successfully and that the output matches the expected configuration
    assert_successful_exit_code(result, cli_cmd)
    assert expected_config_as_str == result.output.strip()

    # Setup: Write the generated configuration to a file
    config_file = "releaserc.toml"
    example_project_dir.joinpath(config_file).write_text(result.output)

    # Act: Validate that the generated config is a valid configuration for PSR
    cli_cmd = [
        MAIN_PROG_NAME,
        "--noop",
        "--strict",
        "-c",
        config_file,
        VERSION_SUBCMD,
        "--print",
    ]
    result = run_cli(cli_cmd[1:])

    # Evaluate: Check that the version command in noop mode ran successfully
    #   which means PSR loaded the configuration successfully
    assert_successful_exit_code(result, cli_cmd)


@pytest.mark.parametrize("args", [("--format", "json"), ("--format", "JSON")])
@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_generate_config_json(
    run_cli: RunCliFn,
    args: tuple[str],
    raw_config_dict: dict[str, Any],
    example_project_dir: ExProjectDir,
):
    # Setup: Generate the expected configuration as a JSON string
    expected_config_as_str = json.dumps(
        {"semantic_release": raw_config_dict}, indent=4
    ).strip()

    # Act: Print the generated configuration to stdout
    cli_cmd = [MAIN_PROG_NAME, GENERATE_CONFIG_SUBCMD, *args]
    result = run_cli(cli_cmd[1:])

    # Evaluate: Check that the command ran successfully and that the output matches the expected configuration
    assert_successful_exit_code(result, cli_cmd)
    assert expected_config_as_str == result.output.strip()

    # Setup: Write the generated configuration to a file
    config_file = "releaserc.json"
    example_project_dir.joinpath(config_file).write_text(result.output)

    # Act: Validate that the generated config is a valid configuration for PSR
    cli_cmd = [
        MAIN_PROG_NAME,
        "--noop",
        "--strict",
        "-c",
        config_file,
        VERSION_SUBCMD,
        "--print",
    ]
    result = run_cli(cli_cmd[1:])

    # Evaluate: Check that the version command in noop mode ran successfully
    #   which means PSR loaded the configuration successfully
    assert_successful_exit_code(result, cli_cmd)


@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_generate_config_pyproject_toml(
    run_cli: RunCliFn,
    raw_config_dict: dict[str, Any],
    example_pyproject_toml: Path,
):
    # Setup: Generate the expected configuration as a TOML string according to PEP 518
    expected_config_as_str = tomlkit.dumps(
        {"tool": {"semantic_release": raw_config_dict}}
    ).strip()

    # Setup: Remove any current configuration from pyproject.toml
    pyproject_config = tomlkit.loads(example_pyproject_toml.read_text(encoding="utf-8"))
    pyproject_config.get("tool", {}).pop("semantic_release", None)
    example_pyproject_toml.write_text(tomlkit.dumps(pyproject_config))

    # Act: Print the generated configuration to stdout
    cli_cmd = [
        MAIN_PROG_NAME,
        GENERATE_CONFIG_SUBCMD,
        "--format",
        "toml",
        "--pyproject",
    ]
    result = run_cli(cli_cmd[1:])

    # Evaluate: Check that the command ran successfully and that the output matches the expected configuration
    assert_successful_exit_code(result, cli_cmd)
    assert expected_config_as_str == result.output.strip()

    # Setup: Write the generated configuration to a file
    example_pyproject_toml.write_text(
        str.join(
            "\n\n",
            [
                example_pyproject_toml.read_text(encoding="utf-8").strip(),
                result.output,
            ],
        )
    )

    # Act: Validate that the generated config is a valid configuration for PSR
    cli_cmd = [MAIN_PROG_NAME, "--noop", "--strict", VERSION_SUBCMD, "--print"]
    result = run_cli(cli_cmd[1:])

    # Evaluate: Check that the version command in noop mode ran successfully
    #   which means PSR loaded the configuration successfully
    assert_successful_exit_code(result, cli_cmd)



from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING

import click
import tomlkit
from git import GitCommandError, Repo

from semantic_release.changelog.release_history import ReleaseHistory
from semantic_release.cli.changelog_writer import (
    generate_release_notes,
    write_changelog_files,
)
from semantic_release.cli.util import noop_report
from semantic_release.globals import logger
from semantic_release.hvcs.remote_hvcs_base import RemoteHvcsBase

if TYPE_CHECKING:  # pragma: no cover
    from semantic_release.cli.cli_context import CliContextObj


def get_license_name_for_release(tag_name: str, project_root: Path) -> str:
    # Retrieve the license name at the time of the specific release tag
    project_metadata: dict[str, str] = {}
    curr_dir = Path.cwd().resolve()
    allowed_directories = [
        dir_path
        for dir_path in [curr_dir, *curr_dir.parents]
        if str(project_root) in str(dir_path)
    ]
    for allowed_dir in allowed_directories:
        proj_toml = allowed_dir.joinpath("pyproject.toml")
        with Repo(project_root) as git_repo, suppress(GitCommandError):
            toml_contents = git_repo.git.show(
                f"{tag_name}:{proj_toml.relative_to(project_root)}"
            )
            config_toml = tomlkit.parse(toml_contents)
            project_metadata = config_toml.unwrap().get("project", project_metadata)
            break

    license_cfg = project_metadata.get(
        "license-expression",
        project_metadata.get(
            "license",
            "",
        ),
    )

    if not isinstance(license_cfg, (str, dict)) or license_cfg is None:
        return ""

    return (
        license_cfg.get("text", "")  # type: ignore[attr-defined]
        if isinstance(license_cfg, dict)
        else license_cfg or ""
    )


def post_release_notes(
    release_tag: str,
    release_notes: str,
    prerelease: bool,
    hvcs_client: RemoteHvcsBase,
    noop: bool = False,
) -> None:
    if noop:
        noop_report(
            str.join(
                "\n",
                [
                    f"would have posted the following release notes for tag {release_tag}:",
                    # Escape square brackets to ensure all content is displayed in the console
                    # (i.e. prevent interpretation of ansi escape sequences that is valid markdown)
                    release_notes.replace("[", "\\["),
                ],
            )
        )
        return

    hvcs_client.create_or_update_release(
        release_tag,
        release_notes,
        prerelease=prerelease,
    )


@click.command(
    short_help="Generate a changelog",
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
@click.option(
    "--post-to-release-tag",
    "release_tag",
    default=None,
    help="Post the generated release notes to the remote VCS's release for this tag",
)
@click.pass_obj
def changelog(cli_ctx: CliContextObj, release_tag: str | None) -> None:
    """Generate and optionally publish a changelog for your project"""
    ctx = click.get_current_context()
    runtime = cli_ctx.runtime_ctx
    translator = runtime.version_translator
    hvcs_client = runtime.hvcs_client

    with Repo(str(runtime.repo_dir)) as git_repo:
        release_history = ReleaseHistory.from_git_history(
            repo=git_repo,
            translator=translator,
            commit_parser=runtime.commit_parser,
            exclude_commit_patterns=runtime.changelog_excluded_commit_patterns,
        )

    write_changelog_files(
        runtime_ctx=runtime,
        release_history=release_history,
        hvcs_client=hvcs_client,
        noop=runtime.global_cli_options.noop,
    )

    if not release_tag:
        return

    if not isinstance(hvcs_client, RemoteHvcsBase):
        click.echo(
            "Remote does not support releases. Skipping release notes update...",
            err=True,
        )
        return

    if not (version := translator.from_tag(release_tag)):
        click.echo(
            str.join(
                " ",
                [
                    f"Tag {release_tag!r} does not match the tag format",
                    repr(translator.tag_format),
                ],
            ),
            err=True,
        )
        ctx.exit(1)

    try:
        release = release_history.released[version]
    except KeyError:
        click.echo(f"tag {release_tag} not in release history", err=True)
        ctx.exit(2)

    release_notes = generate_release_notes(
        hvcs_client,
        release,
        runtime.template_dir,
        release_history,
        style=runtime.changelog_style,
        mask_initial_release=runtime.changelog_mask_initial_release,
        license_name=get_license_name_for_release(
            tag_name=release_tag,
            project_root=runtime.repo_dir,
        ),
    )

    try:
        post_release_notes(
            release_tag=release_tag,
            release_notes=release_notes,
            prerelease=version.is_prerelease,
            hvcs_client=hvcs_client,
            noop=runtime.global_cli_options.noop,
        )
    except Exception as e:  # noqa: BLE001 # TODO: catch specific exceptions
        logger.exception(e)
        click.echo("Failed to post release notes to remote", err=True)
        ctx.exit(1)
